@@ -20,91 +20,13 @@
  * Inc., 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
  */
 
-#include <Python.h>
-#include <tox/tox.h>
+#include "core.h"
+#include "util.h"
 
-PyObject* ToxCoreError;
+extern PyObject* ToxOpError;
 
-#define CHECK_TOX(self)                                        \
-  if ((self)->tox == NULL) {                                   \
-    PyErr_SetString(ToxCoreError, "toxcore object killed.");   \
-    return NULL;                                               \
-  }
-
-#if PY_MAJOR_VERSION < 3
-  #define PYSTRING_FromString PyString_FromString
-  #define PYSTRING_FromStringAndSize PyString_FromStringAndSize
-  #define PYSTRING_Check PyString_Check
-#else
-  #define PYSTRING_FromString PyUnicode_FromString
-  #define PYSTRING_FromStringAndSize PyBytes_FromStringAndSize
-  #define PYSTRING_Check PyUnicode_Check
-#endif
-
-static void bytes_to_hex_string(uint8_t* digest, int length,
-    uint8_t* hex_digest)
-{
-  hex_digest[2 * length] = 0;
-
-  int i, j;
-  for(i = j = 0; i < length; ++i) {
-    char c;
-    c = (digest[i] >> 4) & 0xf;
-    c = (c > 9) ? c + 'A'- 10 : c + '0';
-    hex_digest[j++] = c;
-    c = (digest[i] & 0xf);
-    c = (c > 9) ? c + 'A' - 10 : c + '0';
-    hex_digest[j++] = c;
-  }
-}
-
-static int hex_char_to_int(char c)
-{
-  int val = 0;
-  if (c >= '0' && c <= '9') {
-    val = c - '0';
-  } else if(c >= 'A' && c <= 'F') {
-    val = c - 'A' + 10;
-  } else if(c >= 'a' && c <= 'f') {
-    val = c - 'a' + 10;
-  } else {
-    val = 0;
-  }
-  return val;
-}
-
-static void hex_string_to_bytes(uint8_t* hexstr, int length, uint8_t* bytes)
-{
-  int i;
-  for (i = 0; i < length; ++i) {
-    bytes[i] = (hex_char_to_int(hexstr[2 * i]) << 4)
-             | (hex_char_to_int(hexstr[2 * i + 1]));
-  }
-}
-
-void PyStringUnicode_AsStringAnsSize(PyObject* object, char** str,
-    Py_ssize_t* len)
-{
-#if PY_MAJOR_VERSION < 3
-    PyString_AsStringAndSize(object, str, len);
-#else
-# if PY_MINOR_VERSION == 2
-    *str = PyUnicode_AS_DATA(object);
-    *len = PyUnicode_GET_DATA_SIZE(object);
-# else
-    *str = PyUnicode_AsUTF8AndSize(object, len);
-# endif
-#endif
-}
-
-/* core.Tox definition */
-typedef struct {
-  PyObject_HEAD
-  Tox* tox;
-} ToxCore;
-
-static void callback_friend_request(uint8_t* public_key, uint8_t* data,
-    uint16_t length, void* self)
+static void callback_friend_request(Tox* tox, uint8_t* public_key,
+    uint8_t* data, uint16_t length, void* self)
 {
   uint8_t buf[TOX_CLIENT_ID_SIZE * 2 + 1];
   memset(buf, 0, TOX_CLIENT_ID_SIZE * 2 + 1);
@@ -112,59 +34,66 @@ static void callback_friend_request(uint8_t* public_key, uint8_t* data,
   bytes_to_hex_string(public_key, TOX_CLIENT_ID_SIZE, buf);
 
   PyObject_CallMethod((PyObject*)self, "on_friend_request", "ss#", buf, data,
-      length > 0? length - 1: 0);
+      length);
 }
 
-static void callback_friend_message(Tox *tox, int friendnumber,
+static void callback_friend_message(Tox *tox, int32_t friendnumber,
     uint8_t* message, uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_friend_message", "is#", friendnumber,
-      message, length > 0? length - 1: 0);
+      message, length);
 }
 
-static void callback_action(Tox *tox, int friendnumber, uint8_t* action,
+static void callback_action(Tox *tox, int32_t friendnumber, uint8_t* action,
     uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_friend_action", "is#", friendnumber,
-      action, length > 0? length - 1: 0);
+      action, length);
 }
 
-static void callback_name_change(Tox *tox, int friendnumber, uint8_t* newname,
-    uint16_t length, void* self)
+static void callback_name_change(Tox *tox, int32_t friendnumber,
+    uint8_t* newname, uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_name_change", "is#", friendnumber,
-      newname, length > 0? length - 1: 0);
+      newname, length);
 }
 
-static void callback_status_message(Tox *tox, int friendnumber,
+static void callback_status_message(Tox *tox, int32_t friendnumber,
     uint8_t *newstatus, uint16_t length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_status_message", "is#", friendnumber,
-      newstatus, length > 0? length - 1: 0);
+      newstatus, length);
 }
 
-static void callback_user_status(Tox *tox, int friendnumber,
-    TOX_USERSTATUS kind, void* self)
+static void callback_user_status(Tox *tox, int32_t friendnumber,
+    uint8_t kind, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_user_status", "ii", friendnumber,
       kind);
 }
 
-static void callback_read_receipt(Tox *tox, int friendnumber, uint32_t receipt,
-    void* self)
+static void callback_typing_change(Tox *tox, int32_t friendnumber,
+    uint8_t is_typing, void* self)
+{
+  PyObject_CallMethod((PyObject*)self, "on_typing_change", "iO", friendnumber,
+      PyBool_FromLong(is_typing));
+}
+
+static void callback_read_receipt(Tox *tox, int32_t friendnumber,
+    uint32_t receipt, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_read_receipt", "ii", friendnumber,
       receipt);
 }
 
-static void callback_connection_status(Tox *tox, int friendnumber,
+static void callback_connection_status(Tox *tox, int32_t friendnumber,
     uint8_t status, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_connection_status", "iO",
       friendnumber, PyBool_FromLong(status));
 }
 
-static void callback_group_invite(Tox *tox, int friendnumber,
+static void callback_group_invite(Tox *tox, int32_t friendnumber,
     uint8_t* group_public_key, void *self)
 {
   uint8_t gpk[TOX_CLIENT_ID_SIZE * 2 + 1];
@@ -179,14 +108,14 @@ static void callback_group_message(Tox *tox, int groupid,
     int friendgroupid, uint8_t* message, uint16_t length, void *self)
 {
   PyObject_CallMethod((PyObject*)self, "on_group_message", "iis#", groupid,
-      friendgroupid, message, length > 0? length - 1: 0);
+      friendgroupid, message, length);
 }
 
 static void callback_group_action(Tox *tox, int groupid,
     int friendgroupid, uint8_t* action, uint16_t length, void *self)
 {
   PyObject_CallMethod((PyObject*)self, "on_group_action", "iis#", groupid,
-      friendgroupid, action, length > 0? length - 1: 0);
+      friendgroupid, action, length);
 }
 
 static void callback_group_namelist_change(Tox *tox, int groupid,
@@ -196,16 +125,16 @@ static void callback_group_namelist_change(Tox *tox, int groupid,
       groupid, peernumber, change);
 }
 
-static void callback_file_send_request(Tox *m, int friendnumber,
+static void callback_file_send_request(Tox *m, int32_t friendnumber,
     uint8_t filenumber, uint64_t filesize, uint8_t* filename,
     uint16_t filename_length, void* self)
 {
   PyObject_CallMethod((PyObject*)self, "on_file_send_request", "iiKs#",
       friendnumber, filenumber, filesize, filename,
-      filename_length > 0? filename_length - 1: 0);
+      filename_length);
 }
 
-static void callback_file_control(Tox *m, int friendnumber,
+static void callback_file_control(Tox *m, int32_t friendnumber,
     uint8_t receive_send, uint8_t filenumber, uint8_t control_type,
     uint8_t* data, uint16_t length, void* self)
 {
@@ -218,7 +147,7 @@ static void callback_file_control(Tox *m, int friendnumber,
 #endif
 }
 
-static void callback_file_data(Tox *m, int friendnumber, uint8_t filenumber,
+static void callback_file_data(Tox *m, int32_t friendnumber, uint8_t filenumber,
     uint8_t* data, uint16_t length, void* self)
 {
 #if PY_MAJOR_VERSION < 3
@@ -253,7 +182,7 @@ static int init_helper(ToxCore* self, PyObject* args)
 
       tox = tox_new(0);
       if (tox == NULL) {
-        PyErr_SetString(ToxCoreError, "failed to initialize toxcore");
+        PyErr_SetString(ToxOpError, "failed to initialize toxcore");
         return -1;
       }
   }
@@ -264,6 +193,7 @@ static int init_helper(ToxCore* self, PyObject* args)
   tox_callback_name_change(tox, callback_name_change, self);
   tox_callback_status_message(tox, callback_status_message, self);
   tox_callback_user_status(tox, callback_user_status, self);
+  tox_callback_typing_change(tox, callback_typing_change, self);
   tox_callback_read_receipt(tox, callback_read_receipt, self);
   tox_callback_connection_status(tox, callback_connection_status, self);
   tox_callback_group_invite(tox, callback_group_invite, self);
@@ -351,35 +281,35 @@ ToxCore_add_friend(ToxCore* self, PyObject* args)
   uint8_t pk[TOX_FRIEND_ADDRESS_SIZE];
   hex_string_to_bytes(address, TOX_FRIEND_ADDRESS_SIZE, pk);
 
-  int ret = tox_add_friend(self->tox, pk, data, data_length + 1);
+  int ret = tox_add_friend(self->tox, pk, data, data_length);
   int success = 0;
 
   switch (ret) {
   case TOX_FAERR_TOOLONG:
-    PyErr_SetString(ToxCoreError, "message too long");
+    PyErr_SetString(ToxOpError, "message too long");
     break;
   case TOX_FAERR_NOMESSAGE:
-    PyErr_SetString(ToxCoreError, "no message specified");
+    PyErr_SetString(ToxOpError, "no message specified");
     break;
   case TOX_FAERR_OWNKEY:
-    PyErr_SetString(ToxCoreError, "user's own key");
+    PyErr_SetString(ToxOpError, "user's own key");
     break;
   case TOX_FAERR_ALREADYSENT:
-    PyErr_SetString(ToxCoreError, "friend request already sent or already "
+    PyErr_SetString(ToxOpError, "friend request already sent or already "
         "a friend");
     break;
   case TOX_FAERR_UNKNOWN:
-    PyErr_SetString(ToxCoreError, "unknown error");
+    PyErr_SetString(ToxOpError, "unknown error");
     break;
   case TOX_FAERR_BADCHECKSUM:
-    PyErr_SetString(ToxCoreError, "bad checksum in address");
+    PyErr_SetString(ToxOpError, "bad checksum in address");
     break;
   case TOX_FAERR_SETNEWNOSPAM:
-    PyErr_SetString(ToxCoreError, "the friend was already there but the "
+    PyErr_SetString(ToxOpError, "the friend was already there but the "
         "nospam was different");
     break;
   case TOX_FAERR_NOMEM:
-    PyErr_SetString(ToxCoreError, "increasing the friend list size fails");
+    PyErr_SetString(ToxOpError, "increasing the friend list size fails");
     break;
   default:
     success = 1;
@@ -409,7 +339,7 @@ ToxCore_add_friend_norequest(ToxCore* self, PyObject* args)
 
   int res = tox_add_friend_norequest(self->tox, pk);
   if (res == -1) {
-    PyErr_SetString(ToxCoreError, "failed to add friend");
+    PyErr_SetString(ToxOpError, "failed to add friend");
     return NULL;
   }
 
@@ -431,9 +361,9 @@ ToxCore_get_friend_id(ToxCore* self, PyObject* args)
   uint8_t pk[TOX_CLIENT_ID_SIZE];
   hex_string_to_bytes(id, TOX_CLIENT_ID_SIZE, pk);
 
-  int ret = tox_get_friend_id(self->tox, pk);
+  int ret = tox_get_friend_number(self->tox, pk);
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "no such friend");
+    PyErr_SetString(ToxOpError, "no such friend");
     return NULL;
   }
 
@@ -451,13 +381,13 @@ ToxCore_get_client_id(ToxCore* self, PyObject* args)
   uint8_t hex[TOX_CLIENT_ID_SIZE * 2 + 1];
   memset(hex, 0, TOX_CLIENT_ID_SIZE * 2 + 1);
 
-  int friendid = 0;
+  int friend_num = 0;
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  tox_get_client_id(self->tox, friendid, pk);
+  tox_get_client_id(self->tox, friend_num, pk);
   bytes_to_hex_string(pk, TOX_CLIENT_ID_SIZE, hex);
 
   return PYSTRING_FromString((const char*)hex);
@@ -468,14 +398,14 @@ ToxCore_delfriend(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  if (tox_del_friend(self->tox, friendid) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to delete friend");
+  if (tox_del_friend(self->tox, friend_num) == -1) {
+    PyErr_SetString(ToxOpError, "failed to delete friend");
     return NULL;
   }
 
@@ -487,15 +417,15 @@ ToxCore_get_friend_connection_status(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  int ret = tox_get_friend_connection_status(self->tox, friendid);
+  int ret = tox_get_friend_connection_status(self->tox, friend_num);
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get connection status");
+    PyErr_SetString(ToxOpError, "failed to get connection status");
     return NULL;
   }
 
@@ -507,13 +437,13 @@ ToxCore_friend_exists(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  int ret = tox_friend_exists(self->tox, friendid);
+  int ret = tox_friend_exists(self->tox, friend_num);
 
   return PyBool_FromLong(ret);
 }
@@ -523,17 +453,17 @@ ToxCore_send_message(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   int length = 0;
   uint8_t* message = NULL;
 
-  if (!PyArg_ParseTuple(args, "is#", &friendid, &message, &length)) {
+  if (!PyArg_ParseTuple(args, "is#", &friend_num, &message, &length)) {
     return NULL;
   }
 
-  uint32_t ret = tox_send_message(self->tox, friendid, message, length + 1);
+  uint32_t ret = tox_send_message(self->tox, friend_num, message, length);
   if (ret == 0) {
-    PyErr_SetString(ToxCoreError, "failed to send message");
+    PyErr_SetString(ToxOpError, "failed to send message");
     return NULL;
   }
 
@@ -545,19 +475,19 @@ ToxCore_send_message_withid(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   int length = 0;
   uint32_t id = 0;
   uint8_t* message = NULL;
 
-  if (!PyArg_ParseTuple(args, "iIs#", &friendid, &id, &message, &length)) {
+  if (!PyArg_ParseTuple(args, "iIs#", &friend_num, &id, &message, &length)) {
     return NULL;
   }
 
-  uint32_t ret = tox_send_message_withid(self->tox, friendid, id,message,
-      length + 1);
+  uint32_t ret = tox_send_message_withid(self->tox, friend_num, id,message,
+      length);
   if (ret == 0) {
-    PyErr_SetString(ToxCoreError, "failed to send message with id");
+    PyErr_SetString(ToxOpError, "failed to send message with id");
     return NULL;
   }
 
@@ -569,17 +499,17 @@ ToxCore_send_action(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   int length = 0;
   uint8_t* action = NULL;
 
-  if (!PyArg_ParseTuple(args, "is#", &friendid, &action, &length)) {
+  if (!PyArg_ParseTuple(args, "is#", &friend_num, &action, &length)) {
     return NULL;
   }
 
-  uint32_t ret = tox_send_action(self->tox, friendid, action, length + 1);
+  uint32_t ret = tox_send_action(self->tox, friend_num, action, length);
   if (ret == 0) {
-    PyErr_SetString(ToxCoreError, "failed to send action");
+    PyErr_SetString(ToxOpError, "failed to send action");
     return NULL;
   }
 
@@ -591,19 +521,19 @@ ToxCore_send_action_withid(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   int length = 0;
   uint32_t id = 0;
   uint8_t* action = NULL;
 
-  if (!PyArg_ParseTuple(args, "iIs#", &friendid, &id, &action, &length)) {
+  if (!PyArg_ParseTuple(args, "iIs#", &friend_num, &id, &action, &length)) {
     return NULL;
   }
 
-  uint32_t ret = tox_send_action_withid(self->tox, friendid, id, action,
-      length + 1);
+  uint32_t ret = tox_send_action_withid(self->tox, friend_num, id, action,
+      length);
   if (ret == 0) {
-    PyErr_SetString(ToxCoreError, "failed to send action with id");
+    PyErr_SetString(ToxOpError, "failed to send action with id");
     return NULL;
   }
 
@@ -622,8 +552,8 @@ ToxCore_set_name(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  if (tox_set_name(self->tox, name, length + 1) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to set_name");
+  if (tox_set_name(self->tox, name, length) == -1) {
+    PyErr_SetString(ToxOpError, "failed to set_name");
     return NULL;
   }
 
@@ -638,35 +568,69 @@ ToxCore_get_self_name(ToxCore* self, PyObject* args)
   uint8_t buf[TOX_MAX_NAME_LENGTH];
   memset(buf, 0, TOX_MAX_NAME_LENGTH);
 
-  if (tox_get_self_name(self->tox, buf, TOX_MAX_NAME_LENGTH) == 0) {
-    PyErr_SetString(ToxCoreError, "failed to get self name");
+  if (tox_get_self_name(self->tox, buf) == 0) {
+    PyErr_SetString(ToxOpError, "failed to get self name");
     return NULL;
   }
 
   return PYSTRING_FromString((const char*)buf);
 }
 
+static PyObject*
+ToxCore_get_self_name_size(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int ret = tox_get_self_name_size(self->tox);
+  if (ret == -1) {
+    PyErr_SetString(ToxOpError, "failed to get self name size");
+    return NULL;
+  }
+
+  return PyLong_FromUnsignedLong(ret);
+}
+
 static PyObject*
 ToxCore_get_name(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   uint8_t buf[TOX_MAX_NAME_LENGTH];
   memset(buf, 0, TOX_MAX_NAME_LENGTH);
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  if (tox_get_name(self->tox, friendid, buf) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get name");
+  if (tox_get_name(self->tox, friend_num, buf) == -1) {
+    PyErr_SetString(ToxOpError, "failed to get name");
     return NULL;
   }
 
   return PYSTRING_FromString((const char*)buf);
 }
 
+static PyObject*
+ToxCore_get_name_size(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int friend_num = 0;
+
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
+    return NULL;
+  }
+
+  int ret = tox_get_name_size(self->tox, friend_num);
+  if (ret == -1) {
+    PyErr_SetString(ToxOpError, "failed to get name size");
+    return NULL;
+  }
+
+  return PyLong_FromUnsignedLong(ret);
+}
+
 static PyObject*
 ToxCore_set_status_message(ToxCore* self, PyObject* args)
 {
@@ -678,8 +642,8 @@ ToxCore_set_status_message(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  if (tox_set_status_message(self->tox, message, length + 1) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to set status_message");
+  if (tox_set_status_message(self->tox, message, length) == -1) {
+    PyErr_SetString(ToxOpError, "failed to set status_message");
     return NULL;
   }
 
@@ -698,7 +662,7 @@ ToxCore_set_user_status(ToxCore* self, PyObject* args)
   }
 
   if (tox_set_user_status(self->tox, status) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to set status");
+    PyErr_SetString(ToxOpError, "failed to set status");
     return NULL;
   }
 
@@ -711,13 +675,13 @@ ToxCore_get_status_message_size(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  int friend_num = 0;
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  int ret = tox_get_status_message_size(self->tox, friendid);
-  return PyLong_FromLong(ret - 1);
+  int ret = tox_get_status_message_size(self->tox, friend_num);
+  return PyLong_FromLong(ret);
 }
 
 static PyObject*
@@ -726,19 +690,19 @@ ToxCore_get_status_message(ToxCore* self, PyObject* args)
   CHECK_TOX(self);
 
   uint8_t buf[TOX_MAX_STATUSMESSAGE_LENGTH];
-  int friendid = 0;
+  int friend_num = 0;
 
   memset(buf, 0, TOX_MAX_STATUSMESSAGE_LENGTH);
 
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  int ret = tox_get_status_message(self->tox, friendid, buf,
+  int ret = tox_get_status_message(self->tox, friend_num, buf,
       TOX_MAX_STATUSMESSAGE_LENGTH);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get status_message");
+    PyErr_SetString(ToxOpError, "failed to get status_message");
     return NULL;
   }
 
@@ -759,7 +723,7 @@ ToxCore_get_self_status_message(ToxCore* self, PyObject* args)
       TOX_MAX_STATUSMESSAGE_LENGTH);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get self status_message");
+    PyErr_SetString(ToxOpError, "failed to get self status_message");
     return NULL;
   }
 
@@ -768,17 +732,31 @@ ToxCore_get_self_status_message(ToxCore* self, PyObject* args)
   return PYSTRING_FromString((const char*)buf);
 }
 
+static PyObject*
+ToxCore_get_self_status_message_size(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int ret = tox_get_self_status_message_size(self->tox);
+  if (ret == -1) {
+    PyErr_SetString(ToxOpError, "failed to get self status_message size");
+    return NULL;
+  }
+
+  return PyLong_FromLong(ret);
+}
+
 static PyObject*
 ToxCore_get_user_status(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
-  if (!PyArg_ParseTuple(args, "i", &friendid)) {
+  int friend_num = 0;
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
     return NULL;
   }
 
-  int status = tox_get_user_status(self->tox, friendid);
+  int status = tox_get_user_status(self->tox, friend_num);
 
   return PyLong_FromLong(status);
 }
@@ -792,19 +770,77 @@ ToxCore_get_self_user_status(ToxCore* self, PyObject* args)
   return PyLong_FromLong(status);
 }
 
+static PyObject*
+ToxCore_get_last_online(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int friend_num = 0;
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
+    return NULL;
+  }
+
+  uint64_t status = tox_get_last_online(self->tox, friend_num);
+
+  if (status == 0) {
+    Py_RETURN_NONE;
+  }
+
+  PyObject* datetime = PyImport_ImportModule("datetime");
+  PyObject* datetimeClass = PyObject_GetAttrString(datetime, "datetime");
+  PyObject* ret = PyObject_CallMethod(datetimeClass, "fromtimestamp", "K",
+      status);
+
+  return ret;
+}
+
+static PyObject*
+ToxCore_set_user_is_typing(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int friend_num = 0;
+  int is_typing = 0;
+
+  if (!PyArg_ParseTuple(args, "ii", &friend_num, &is_typing)) {
+    return NULL;
+  }
+
+  int status = tox_set_user_is_typing(self->tox, friend_num, is_typing);
+  return PyLong_FromLong(status);
+}
+
+static PyObject*
+ToxCore_get_is_typing(ToxCore* self, PyObject* args)
+{
+  CHECK_TOX(self);
+
+  int friend_num = 0;
+
+  if (!PyArg_ParseTuple(args, "i", &friend_num)) {
+    return NULL;
+  }
+
+  if (tox_get_is_typing(self->tox, friend_num)) {
+    Py_RETURN_TRUE;
+  } else {
+    Py_RETURN_FALSE;
+  }
+}
+
 static PyObject*
 ToxCore_set_send_receipts(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendid = 0;
+  int friend_num = 0;
   int yesno = 0;
 
-  if (!PyArg_ParseTuple(args, "ii", &friendid, &yesno)) {
+  if (!PyArg_ParseTuple(args, "ii", &friend_num, &yesno)) {
     return NULL;
   }
 
-  tox_set_sends_receipts(self->tox, friendid, yesno);
+  tox_set_sends_receipts(self->tox, friend_num, yesno);
 
   Py_RETURN_NONE;
 }
@@ -858,7 +894,7 @@ ToxCore_add_groupchat(ToxCore* self, PyObject* args)
 
   int ret = tox_add_groupchat(self->tox);
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to add groupchat");
+    PyErr_SetString(ToxOpError, "failed to add groupchat");
   }
 
   return PyLong_FromLong(ret);
@@ -876,7 +912,7 @@ ToxCore_del_groupchat(ToxCore* self, PyObject* args)
   }
 
   if (tox_del_groupchat(self->tox, groupid) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to del groupchat");
+    PyErr_SetString(ToxOpError, "failed to del groupchat");
   }
 
   Py_RETURN_NONE;
@@ -898,7 +934,7 @@ ToxCore_group_peername(ToxCore* self, PyObject* args)
 
   int ret = tox_group_peername(self->tox, groupid, peernumber, buf);
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get group peername");
+    PyErr_SetString(ToxOpError, "failed to get group peername");
   }
 
   return PYSTRING_FromString((const char*)buf);
@@ -909,14 +945,14 @@ ToxCore_invite_friend(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   int groupid = 0;
   if (!PyArg_ParseTuple(args, "ii", &friendnumber, &groupid)) {
     return NULL;
   }
 
   if (tox_invite_friend(self->tox, friendnumber, groupid) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to invite friend");
+    PyErr_SetString(ToxOpError, "failed to invite friend");
   }
 
   Py_RETURN_NONE;
@@ -929,7 +965,7 @@ ToxCore_join_groupchat(ToxCore* self, PyObject* args)
 
   uint8_t* public_key = NULL;
   int length = 0;
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   uint8_t pk[TOX_CLIENT_ID_SIZE];
 
   if (!PyArg_ParseTuple(args, "is#", &friendnumber, &public_key, &length)) {
@@ -940,7 +976,7 @@ ToxCore_join_groupchat(ToxCore* self, PyObject* args)
 
   int ret = tox_join_groupchat(self->tox, friendnumber, pk);
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "failed to join group chat");
+    PyErr_SetString(ToxOpError, "failed to join group chat");
   }
 
   return PyLong_FromLong(ret);
@@ -959,8 +995,8 @@ ToxCore_group_message_send(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  if (tox_group_message_send(self->tox, groupid, message, length + 1) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to send group message");
+  if (tox_group_message_send(self->tox, groupid, message, length) == -1) {
+    PyErr_SetString(ToxOpError, "failed to send group message");
   }
 
   Py_RETURN_NONE;
@@ -979,8 +1015,8 @@ ToxCore_group_action_send(ToxCore* self, PyObject* args)
     return NULL;
   }
 
-  if (tox_group_action_send(self->tox, groupid, action, length + 1) == -1) {
-    PyErr_SetString(ToxCoreError, "failed to send group action");
+  if (tox_group_action_send(self->tox, groupid, action, length) == -1) {
+    PyErr_SetString(ToxOpError, "failed to send group action");
   }
 
   Py_RETURN_NONE;
@@ -1013,12 +1049,13 @@ ToxCore_group_get_names(ToxCore* self, PyObject* args)
   }
 
   int n = tox_group_number_peers(self->tox, groupid);
+  uint16_t* lengths = (uint16_t*)malloc(sizeof(uint16_t) * n);
   uint8_t (*names)[TOX_MAX_NAME_LENGTH] = (uint8_t(*)[TOX_MAX_NAME_LENGTH])
-    malloc(sizeof(uint8_t*) * n * TOX_MAX_NAME_LENGTH);
+    malloc(sizeof(uint8_t) * n * TOX_MAX_NAME_LENGTH);
 
-  int n2 = tox_group_get_names(self->tox, groupid, names, n);
+  int n2 = tox_group_get_names(self->tox, groupid, names, lengths, n);
   if (n2 == -1) {
-    PyErr_SetString(ToxCoreError, "failed to get group member names");
+    PyErr_SetString(ToxOpError, "failed to get group member names");
     return NULL;
   }
 
@@ -1029,10 +1066,12 @@ ToxCore_group_get_names(ToxCore* self, PyObject* args)
 
   int i = 0;
   for (i = 0; i < n2; ++i) {
-    PyList_Append(list, PYSTRING_FromString((const char*)names[i]));
+    PyList_Append(list,
+        PYSTRING_FromStringAndSize((const char*)names[i], lengths[i]));
   }
 
   free(names);
+  free(lengths);
 
   return list;
 }
@@ -1075,7 +1114,7 @@ ToxCore_new_filesender(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   uint64_t filesize = 0;
   uint8_t* filename = 0;
   int filename_length = 0;
@@ -1086,10 +1125,10 @@ ToxCore_new_filesender(ToxCore* self, PyObject* args)
   }
 
   int ret = tox_new_file_sender(self->tox, friendnumber, filesize, filename,
-      filename_length + 1);
+      filename_length);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "tox_new_file_sender() failed");
+    PyErr_SetString(ToxOpError, "tox_new_file_sender() failed");
     return NULL;
   }
 
@@ -1101,7 +1140,7 @@ ToxCore_file_sendcontrol(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   uint8_t send_receive = 0;
   int filenumber = 0;
   uint8_t message_id = 0;
@@ -1118,7 +1157,7 @@ ToxCore_file_sendcontrol(ToxCore* self, PyObject* args)
       filenumber, message_id, data, data_length);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "tox_file_send_control() failed");
+    PyErr_SetString(ToxOpError, "tox_file_send_control() failed");
     return NULL;
   }
 
@@ -1130,7 +1169,7 @@ ToxCore_file_senddata(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   int filenumber = 0;
   uint8_t* data = NULL;
   uint32_t data_length = 0;
@@ -1145,7 +1184,7 @@ ToxCore_file_senddata(ToxCore* self, PyObject* args)
       data_length);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "tox_file_send_data() failed");
+    PyErr_SetString(ToxOpError, "tox_file_send_data() failed");
     return NULL;
   }
 
@@ -1157,7 +1196,7 @@ ToxCore_file_data_size(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   if (!PyArg_ParseTuple(args, "i", &friendnumber)) {
     return NULL;
   }
@@ -1165,7 +1204,7 @@ ToxCore_file_data_size(ToxCore* self, PyObject* args)
   int ret = tox_file_data_size(self->tox, friendnumber);
 
   if (ret == -1) {
-    PyErr_SetString(ToxCoreError, "tox_file_data_size() failed");
+    PyErr_SetString(ToxOpError, "tox_file_data_size() failed");
     return NULL;
   }
 
@@ -1177,7 +1216,7 @@ ToxCore_file_data_remaining(ToxCore* self, PyObject* args)
 {
   CHECK_TOX(self);
 
-  int friendnumber = 0;
+  int32_t friendnumber = 0;
   int filenumber = 0;
   int send_receive = 0;
 
@@ -1191,7 +1230,7 @@ ToxCore_file_data_remaining(ToxCore* self, PyObject* args)
       send_receive);
 
   if (!ret) {
-    PyErr_SetString(ToxCoreError, "tox_file_data_remaining() failed");
+    PyErr_SetString(ToxOpError, "tox_file_data_remaining() failed");
     return NULL;
   }
 
@@ -1221,7 +1260,7 @@ ToxCore_bootstrap_from_address(ToxCore* self, PyObject* args)
       htons(port), pk);
 
   if (!ret) {
-    PyErr_SetString(ToxCoreError, "failed to resolve address");
+    PyErr_SetString(ToxOpError, "failed to resolve address");
     return NULL;
   }
 
@@ -1290,7 +1329,7 @@ ToxCore_save(ToxCore* self, PyObject* args)
   if (PYSTRING_Check(keyobj)) {
     PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
   } else if (keyobj != Py_None) {
-    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    PyErr_SetString(ToxOpError, "invalid passphrase type");
     return NULL;
   }
 
@@ -1308,7 +1347,7 @@ ToxCore_save(ToxCore* self, PyObject* args)
 
   if (key) {
     if (tox_save_encrypted(self->tox, buf, key, key_len) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_save_encrypted(): failed");
+      PyErr_SetString(ToxOpError, "tox_save_encrypted(): failed");
       free(buf);
       return NULL;
     }
@@ -1316,7 +1355,7 @@ ToxCore_save(ToxCore* self, PyObject* args)
     tox_save(self->tox, buf);
   }
 
-  PyObject* res = PYSTRING_FromStringAndSize((const char*)buf, size);
+  PyObject* res = PYBYTES_FromStringAndSize((const char*)buf, size);
   free(buf);
 
   return res;
@@ -1339,7 +1378,7 @@ ToxCore_save_to_file(ToxCore* self, PyObject* args)
   if (PYSTRING_Check(keyobj)) {
     PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
   } else if (keyobj != Py_None) {
-    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    PyErr_SetString(ToxOpError, "invalid passphrase type");
     return NULL;
   }
 
@@ -1358,7 +1397,7 @@ ToxCore_save_to_file(ToxCore* self, PyObject* args)
 
   if (key) {
     if (tox_save_encrypted(self->tox, buf, key, key_len) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_save_encrypted(): failed");
+      PyErr_SetString(ToxOpError, "tox_save_encrypted(): failed");
       free(buf);
       return NULL;
     }
@@ -1369,7 +1408,7 @@ ToxCore_save_to_file(ToxCore* self, PyObject* args)
   FILE* fp = fopen(filename, "w");
 
   if (fp == NULL) {
-    PyErr_SetString(ToxCoreError, "fopen(): can't open file for saving");
+    PyErr_SetString(ToxOpError, "fopen(): can't open file for saving");
     return NULL;
   }
 
@@ -1393,25 +1432,25 @@ ToxCore_load(ToxCore* self, PyObject* args)
   Py_ssize_t key_len = 0;
 
   if (!PyArg_ParseTuple(args, "s#|O", &data, &length, &keyobj)) {
-    PyErr_SetString(ToxCoreError, "no data specified");
+    PyErr_SetString(ToxOpError, "no data specified");
     return NULL;
   }
 
   if (PYSTRING_Check(keyobj)) {
     PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
   } else if (keyobj != Py_None) {
-    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    PyErr_SetString(ToxOpError, "invalid passphrase type");
     return NULL;
   }
 
   if (key) {
     if (tox_load_encrypted(self->tox, data, length, key, key_len) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_load_encrypted(): load failed");
+      PyErr_SetString(ToxOpError, "tox_load_encrypted(): load failed");
       return NULL;
     }
   } else {
     if (tox_load(self->tox, data, length) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_load(): load failed");
+      PyErr_SetString(ToxOpError, "tox_load(): load failed");
       return NULL;
     }
   }
@@ -1436,13 +1475,13 @@ ToxCore_load_from_file(ToxCore* self, PyObject* args)
   if (PYSTRING_Check(keyobj)) {
     PyStringUnicode_AsStringAnsSize(keyobj, (char**)&key, &key_len);
   } else if (keyobj != Py_None) {
-    PyErr_SetString(ToxCoreError, "invalid passphrase type");
+    PyErr_SetString(ToxOpError, "invalid passphrase type");
     return NULL;
   }
 
   FILE* fp = fopen(filename, "r");
   if (fp == NULL) {
-    PyErr_SetString(ToxCoreError,
+    PyErr_SetString(ToxOpError,
         "tox_load(): failed to open file for reading");
     return NULL;
   }
@@ -1455,7 +1494,7 @@ ToxCore_load_from_file(ToxCore* self, PyObject* args)
   uint8_t* data = (uint8_t*)malloc(length * sizeof(char));
 
   if (fread(data, length, 1, fp) != 1) {
-    PyErr_SetString(ToxCoreError,
+    PyErr_SetString(ToxOpError,
         "tox_load(): corrupted data file");
     free(data);
     return NULL;
@@ -1463,13 +1502,13 @@ ToxCore_load_from_file(ToxCore* self, PyObject* args)
 
   if (key) {
     if (tox_load_encrypted(self->tox, data, length, key, key_len) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_load_encrypted(): load failed");
+      PyErr_SetString(ToxOpError, "tox_load_encrypted(): load failed");
       free(data);
       return NULL;
     }
   } else {
     if (tox_load(self->tox, data, length) == -1) {
-      PyErr_SetString(ToxCoreError, "tox_load(): load failed");
+      PyErr_SetString(ToxOpError, "tox_load(): load failed");
       free(data);
       return NULL;
     }
@@ -1518,6 +1557,12 @@ PyMethodDef Tox_methods[] = {
     ".. seealso ::\n"
     "    :meth:`.set_user_status`"
   },
+  {
+    "on_typing_change", (PyCFunction)ToxCore_callback_stub, METH_VARARGS,
+    "on_typing_change(friend_number, is_typing)\n"
+    "Callback for receiving friend status changes, default implementation "
+    "does nothing.\n\n"
+  },
   {
     "on_read_receipt", (PyCFunction)ToxCore_callback_stub, METH_VARARGS,
     "on_read_receipt(friend_number, receipt)\n"
@@ -1661,10 +1706,20 @@ PyMethodDef Tox_methods[] = {
     "get_self_name()\n"
     "Get your nickname."
   },
+  {
+    "get_self_name_size", (PyCFunction)ToxCore_get_self_name_size, METH_NOARGS,
+    "get_self_name_size()\n"
+    "Get your nickname string length"
+  },
   {
     "get_name", (PyCFunction)ToxCore_get_name, METH_VARARGS,
     "get_name(friend_number)\n"
-    "Get name of *friend_number*."
+    "Get nickname of *friend_number*."
+  },
+  {
+    "get_name_size", (PyCFunction)ToxCore_get_name_size, METH_VARARGS,
+    "get_name_size(friend_number)\n"
+    "Get nickname length of *friend_number*."
   },
   {
     "set_status_message", (PyCFunction)ToxCore_set_status_message, METH_VARARGS,
@@ -1705,6 +1760,13 @@ PyMethodDef Tox_methods[] = {
     "get_self_status_message()\n"
     "Get status message of yourself."
   },
+  {
+    "get_self_status_message_size",
+    (PyCFunction)ToxCore_get_self_status_message_size,
+    METH_NOARGS,
+    "get_self_status_message_size()\n"
+    "Get status message string length of yourself."
+  },
   {
     "get_user_status", (PyCFunction)ToxCore_get_user_status,
     METH_VARARGS,
@@ -1715,12 +1777,31 @@ PyMethodDef Tox_methods[] = {
   },
   {
     "get_self_user_status", (PyCFunction)ToxCore_get_self_user_status,
-    METH_VARARGS,
+    METH_NOARGS,
     "get_self_user_status()\n"
     "Get user status of youself.\n\n"
     ".. seealso ::\n"
     "    :meth:`.set_user_status`"
   },
+  {
+    "get_last_online", (PyCFunction)ToxCore_get_last_online,
+    METH_VARARGS,
+    "get_last_online(friend_number)\n"
+    "returns datetime.datetime object representing the last time "
+    "*friend_number* was seen online, or None if never seen."
+  },
+  {
+    "set_user_is_typing", (PyCFunction)ToxCore_set_user_is_typing,
+    METH_VARARGS,
+    "set_user_is_typing(friend_number, is_typing)\n"
+    "Set user typing status.\n\n"
+  },
+  {
+    "get_is_typing", (PyCFunction)ToxCore_get_is_typing,
+    METH_VARARGS,
+    "get_is_typing(friend_number)\n"
+    "Return True is user is typing.\n\n"
+  },
   {
     "set_send_receipts", (PyCFunction)ToxCore_set_send_receipts,
     METH_VARARGS,
@@ -1901,7 +1982,7 @@ PyTypeObject ToxCoreType = {
   PyObject_HEAD_INIT(NULL)
   0,                         /*ob_size*/
 #endif
-  "core.Tox",                /*tp_name*/
+  "Tox",                     /*tp_name*/
   sizeof(ToxCore),           /*tp_basicsize*/
   0,                         /*tp_itemsize*/
   (destructor)ToxCore_dealloc, /*tp_dealloc*/
@@ -1942,8 +2023,10 @@ PyTypeObject ToxCoreType = {
 
 void ToxCore_install_dict()
 {
-#define SET(name) \
-  PyDict_SetItemString(dict, #name, PyLong_FromLong(TOX_##name));
+#define SET(name)                                            \
+  PyObject* obj_##name = PyLong_FromLong(TOX_##name);        \
+  PyDict_SetItemString(dict, #name, obj_##name);             \
+  Py_DECREF(obj_##name);
 
   PyObject* dict = PyDict_New();
   SET(FAERR_TOOLONG)
