@@ -21,13 +21,13 @@
  */
 
 #include <Python.h>
-
 #include <stdio.h>
 
-extern PyTypeObject ToxCoreType;
-extern PyObject* ToxCoreError;
+#include "core.h"
 
-extern void ToxCore_install_dict();
+extern PyObject* ToxOpError;
+extern PyTypeObject ToxCoreType;
+extern PyTypeObject ToxAVType;
 
 #if PY_MAJOR_VERSION >= 3
 struct PyModuleDef moduledef = {
@@ -58,8 +58,9 @@ PyMODINIT_FUNC inittox(void)
   }
 
   ToxCore_install_dict();
+  ToxAV_install_dict();
 
-  // Initialize tox.core
+  // Initialize toxcore
   if (PyType_Ready(&ToxCoreType) < 0) {
     fprintf(stderr, "Invalid PyTypeObject `ToxCoreType'\n");
     goto error;
@@ -68,8 +69,17 @@ PyMODINIT_FUNC inittox(void)
   Py_INCREF(&ToxCoreType);
   PyModule_AddObject(m, "Tox", (PyObject*)&ToxCoreType);
 
-  ToxCoreError = PyErr_NewException("tox.OperationFailedError", NULL, NULL);
-  PyModule_AddObject(m, "OperationFailedError", (PyObject*)ToxCoreError);
+  ToxOpError = PyErr_NewException("tox.OperationFailedError", NULL, NULL);
+  PyModule_AddObject(m, "OperationFailedError", (PyObject*)ToxOpError);
+
+  // Initialize toxav
+  if (PyType_Ready(&ToxAVType) < 0) {
+    fprintf(stderr, "Invalid PyTypeObject `ToxAVType'\n");
+    goto error;
+  }
+
+  Py_INCREF(&ToxAVType);
+  PyModule_AddObject(m, "ToxAV", (PyObject*)&ToxAVType);
 
 #if PY_MAJOR_VERSION >= 3
   return m;
